from typing import Callable, Union, List, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from pyfileconf.pipelines.models.collection import PipelineCollection

from pyfileconf.views.object import ObjectView

FunctionOrCollection = Union[Callable, 'PipelineCollection']

StrList = List[str]
ObjectViewOrCollection = Union[ObjectView, 'PipelineCollection']


"""
Default behavior to be run on hooks
"""
from typing import Any, Dict, Sequence, List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from pyfileconf.iterate import IterativeRunner

from pyfileconf.plugin.impl import hookimpl


@hookimpl
def pyfileconf_iter_get_cases(
    config_updates: Sequence[Dict[str, Any]], runner: "IterativeRunner"
) -> List[Tuple[Dict[str, Any], ...]]:
    """
    Collect user passed casses by section path str then call
    itertools.product to produce the cases

    :param config_updates: list of kwarg dictionaries which would normally be provided to .update
    :return: config cases to be run
    """
    from pyfileconf.iterate import get_config_product

    return get_config_product(config_updates)


@hookimpl
def pyfileconf_iter_update_for_case(
    case: Tuple[Dict[str, Any], ...], runner: "IterativeRunner"
):
    """
    Called in PipelineManager.run_iter and IterativeRunner to take the case
    containing all the updates and actually run the updates, before running
    this case.

    :param case: tuple of kwarg dictionaries which would normally be provided to .update
    :param runner: :class:`IterativeRunner` which has been constructed to call iteration
    :return: None
    """
    from pyfileconf import PipelineManager
    from pyfileconf.sectionpath.sectionpath import SectionPath

    for config_dict in case:
        sp_str = config_dict["section_path_str"]
        if runner.base_section_path_str is not None:
            sp_str = SectionPath.join(
                runner.base_section_path_str, sp_str
            ).path_str
        manager = PipelineManager.get_manager_by_section_path_str(sp_str)
        relative_section_path_str = SectionPath(
            ".".join(SectionPath(sp_str)[1:])
        ).path_str
        manager.reset(relative_section_path_str)
        relative_conf_dict = {**config_dict}
        relative_conf_dict["section_path_str"] = relative_section_path_str
        manager.update(**relative_conf_dict)


from functools import partial
from typing import Any, Dict

from pyfileconf.imports.logic.load.func import function_args_as_dict


def apply_config(obj: Any, config: dict) -> None:
    if isinstance(obj, partial):
        apply_config_to_partial(obj, config)
    else:
        apply_config_to_obj(obj, config)


def apply_config_to_obj(obj: Any, config: dict) -> None:
    attributes = dir(obj)
    init_args = function_args_as_dict(obj.__init__)
    relevant_config: Dict[str, Any] = {
        attr: value for attr, value in config.items() if attr in init_args.keys()
    }
    obj.__init__(**relevant_config)


def apply_config_to_partial(part: partial, config: dict) -> None:
    attributes = part.keywords.keys()
    for config_attr, config_item in config.items():
        # Skip irrelevant items
        if config_attr in attributes:
            part.keywords[config_attr] = config_item
